#include "FreeRTOS.h"
#include "GPIOINT.h"
#include "bit.h"
#include "FreeRTOSConfig.h"
#include "portmacro.h"
#include "task.h"
#include "Timer.h"

/**
 * An array of the GPIO interrupts that need to be matched upon an interrupt
 */
static GPIOINT_desc *subscribers[GPIOINT_MAX_SUBSCRIBERS];

void GPIOINT_enable()
{
	Timer_Setup();
	//Clear interrupts for a clean slate
	taskENTER_CRITICAL();
	NVIC_DisableIRQ(EINT3_IRQn);
	LPC_GPIOINT->IO0IntClr = 0xFFFFFFFF;
	LPC_GPIOINT->IO2IntClr = 0xFFFFFFFF;
	NVIC_ClearPendingIRQ(EINT3_IRQn);
	//Enable them
	NVIC_SetPriority(EINT3_IRQn, GPIOINT_NVIC_PRIORITY);
	taskEXIT_CRITICAL();
	NVIC_EnableIRQ(EINT3_IRQn);
}

/**
 * Enables a GPIO rising interrupt
 * @param port The port number (0 or 2)
 * @param The pin number
 *
 * @return 0 on failure, 1 on success
 */
static uint8_t _GPIOINT_enable_rising(uint8_t port, uint8_t pin)
{
	if (port == 0) {
		Bit_on(LPC_GPIOINT->IO0IntEnR, 1 << pin);
		return 1;
	}

	if (port == 2) {
		Bit_on(LPC_GPIOINT->IO2IntEnR, 1 << pin);
		return 1;
	}

	return 0;
}

/**
 * Enables a GPIO falling interrupt
 * @param port The port number (0 or 2)
 * @param The pin number
 *
 * @return 0 on failure, 1 on success
 */
static uint8_t _GPIOINT_enable_falling(uint8_t port, uint8_t pin)
{
	if (port == 0) {
		Bit_on(LPC_GPIOINT->IO0IntEnF, 1 << pin);
		return 1;
	}

	if (port == 2) {
		Bit_on(LPC_GPIOINT->IO2IntEnF, 1 << pin);
		return 1;
	}

	return 0;
}

/**
 * Add a GPIOINT descriptor to a free slot in the subscriptors array
 *
 * @return 0 on failure, 1 on success
 */
static uint8_t _GPIOINT_set_subscriber(GPIOINT_desc *desc)
{
	uint8_t i;
	for(i=0; i < GPIOINT_MAX_SUBSCRIBERS; ++i) {
		if (subscribers[i] == 0) {
			subscribers[i] = desc;
			return 1;
		}
	}
	return 0;
}

uint8_t GPIOINT_subscribe(GPIOINT_desc *desc, uint8_t port, uint8_t pin, GPIOINT_direction direction, unsigned portBASE_TYPE queueSize)
{
	if (port == 0) {
		if (direction == rising) {
			desc->statusRegister = (uint32_t *) &(LPC_GPIOINT->IO0IntStatR);
		} else {
			desc->statusRegister = (uint32_t *) &(LPC_GPIOINT->IO0IntStatF);
		}
		desc->clearRegister = &(LPC_GPIOINT->IO0IntClr);
	} else if (port == 2) {
		if (direction == rising) {
			desc->statusRegister = (uint32_t *) &(LPC_GPIOINT->IO2IntStatR);
		} else {
			desc->statusRegister = (uint32_t *) &(LPC_GPIOINT->IO2IntStatF);
		}
		desc->clearRegister = &(LPC_GPIOINT->IO2IntClr);
	} else {
		return 0;
	}

	desc->statusBitMask = 1 << pin;
	desc->queueHandle = xQueueCreate(queueSize, sizeof(uint32_t));
	_GPIOINT_set_subscriber(desc);

	if (direction == rising) {
		_GPIOINT_enable_rising(port, pin);
	} else {
		_GPIOINT_enable_falling(port, pin);
	}

	return 1;
}

void EINT3_IRQHandler(void)
{
	uint32_t microSeconds;
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	uint8_t i;

	for(i=0; i < GPIOINT_MAX_SUBSCRIBERS; ++i) {
		if (subscribers[i] == 0) {
			break;
		}

		if (Bit_isSet(*(subscribers[i]->statusRegister), subscribers[i]->statusBitMask)) {
			microSeconds = *Timer_usTime;
			xQueueSendToBackFromISR(subscribers[i]->queueHandle, (void *) &microSeconds, &xHigherPriorityTaskWoken);
			Bit_on(*(subscribers[i]->clearRegister), subscribers[i]->statusBitMask);
		}
	}

    portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
}
