#include "UART.h"
#include "LPC17xx.h"
#include "portmacro.h"
#include "bit.h"
#include "semphr.h"
#include "task.h"

static uint8_t UartTxEmpty[] = { 1, 1, 1, 1 };
static xQueueHandle queues[4];
static xSemaphoreHandle uartWriteMutex[4];

uint8_t UART_Setup(uint8_t uartNum, UART_WordLength wordLength, uint32_t bauds, xQueueHandle queue)
{
	static uint8_t bits[] = { 3, 4, 24, 25 };
	if (uartNum > 3) {
		return 0;
	}
	queues[uartNum] = queue;

	uint16_t dl = (25000000 / 16) / bauds;

	Bit_on(LPC_SC->PCONP, 1 << bits[uartNum]);
	NVIC_DisableIRQ(UART0_IRQn + uartNum);
	NVIC_ClearPendingIRQ(UART0_IRQn + uartNum);
	switch(uartNum) {
		case 0:
			LPC_UART0->LCR = wordLength | (1 << 7);
			LPC_UART0->DLL = dl & 0xFF;
			LPC_UART0->DLM = dl >> 8;
			Bit_off(LPC_PINCON->PINSEL0, 0xF0);
			Bit_on(LPC_PINCON->PINSEL0, (1 << 4) | (1 << 6));
			Bit_off(LPC_UART0->LCR, 1 << 7);
			LPC_UART0->FCR = 7;
			NVIC_EnableIRQ(UART0_IRQn);
			LPC_UART0->IER = 7;
			break;
		case 1:
			LPC_UART1->LCR = wordLength | (1 << 7);
			LPC_UART1->DLL = dl & 0xFF;
			LPC_UART1->DLM = dl >> 8;
			Bit_off(LPC_PINCON->PINSEL0, 1 << 31);
			Bit_on(LPC_PINCON->PINSEL0, 1 << 30);
			Bit_off(LPC_PINCON->PINSEL1, 1 << 1);
			Bit_on(LPC_PINCON->PINSEL1, 1);
			Bit_off(LPC_UART1->LCR, 1 << 7);
			LPC_UART1->FCR = 7;
			NVIC_EnableIRQ(UART1_IRQn);
			LPC_UART1->IER = 7;
			break;
		case 2:
			LPC_UART2->LCR = wordLength | (1 << 7);
			LPC_UART2->DLL = dl & 0xFF;
			LPC_UART2->DLM = dl >> 8;
			Bit_off(LPC_PINCON->PINSEL0, 0xF << 20);
			Bit_on(LPC_PINCON->PINSEL0, (1 << 20) | (1 << 22));
			Bit_off(LPC_UART2->LCR, 1 << 7);
			LPC_UART2->FCR = 7;
			NVIC_EnableIRQ(UART2_IRQn);
			LPC_UART2->IER = 7;
			break;
		case 3:
			LPC_UART3->LCR = wordLength | (1 << 7);
			LPC_UART3->DLL = dl & 0xFF;
			LPC_UART3->DLM = dl >> 8;
			Bit_off(LPC_PINCON->PINSEL0, 0xF);
			Bit_on(LPC_PINCON->PINSEL0, (1 << 1) | (1 << 3));
			Bit_off(LPC_UART3->LCR, 1 << 7);
			LPC_UART3->FCR = 7;
			NVIC_EnableIRQ(UART3_IRQn);
			LPC_UART3->IER = 7;
			break;
		default:
			return 0;
	}

	if (!uartWriteMutex[uartNum]) {
		vSemaphoreCreateBinary(uartWriteMutex[uartNum]);
		NVIC_SetPriority(UART0_IRQn + uartNum, UART_NVIC_PRIORITY);
	}

	return 1;
}

void UART0_IRQHandler()
{
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	uint8_t interruptType = (LPC_UART0->IIR >> 1) & 7, character;

	switch(interruptType) {
		case 1: //THRE
			UartTxEmpty[0] = (LPC_UART0->LSR >> 5) & 1;
			break;
		case 2: //Receive data available
			character = LPC_UART0->RBR;
			if (&queues[0]) {
				xQueueSendToBackFromISR(queues[0], &character, &xHigherPriorityTaskWoken);
			}
			break;
		case 3: //Receive line status
			if (LPC_UART0->LSR & (1 << 7)) {
				character = LPC_UART0->RBR;
			}
			break;
		case 4: //Character time out indicator
		default:
			character = LPC_UART0->LSR;
			break;
	}

	portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
}

void UART1_IRQHandler()
{
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	uint8_t interruptType = (LPC_UART1->IIR >> 1) & 7, character;

	switch(interruptType) {
		case 1: //THRE
			UartTxEmpty[1] = (LPC_UART1->LSR >> 5) & 1;
			break;
		case 2: //Receive data available
			character = LPC_UART1->RBR;
			if (&queues[1]) {
				xQueueSendToBackFromISR(queues[1], &character, &xHigherPriorityTaskWoken);
			}
			break;
		case 3: //Receive line status
			if (LPC_UART1->LSR & (1 << 7)) {
				character = LPC_UART1->RBR;
			}
			break;
		case 4: //Character time out indicator
		default:
			character = LPC_UART1->LSR;
			break;
	}

	portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
}

void UART2_IRQHandler()
{
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	uint8_t interruptType = (LPC_UART2->IIR >> 1) & 7, character;

	switch(interruptType) {
		case 1: //THRE
			UartTxEmpty[2] = (LPC_UART2->LSR >> 5) & 1;
			break;
		case 2: //Receive data available
			character = LPC_UART2->RBR;
			if (&queues[2]) {
				xQueueSendToBackFromISR(queues[2], &character, &xHigherPriorityTaskWoken);
			}
			break;
		case 3: //Receive line status
			if (LPC_UART2->LSR & (1 << 7)) {
				character = LPC_UART2->RBR;
			}
			break;
		case 4: //Character time out indicator
		default:
			character = LPC_UART2->LSR;
			break;
	}

	portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
}

void UART3_IRQHandler()
{
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	uint8_t interruptType = (LPC_UART3->IIR >> 1) & 7, character;

	switch(interruptType) {
		case 1: //THRE
			UartTxEmpty[3] = (LPC_UART3->LSR >> 5) & 1;
			break;
		case 2: //Receive data available
			character = LPC_UART3->RBR;
			if (&queues[3]) {
				xQueueSendToBackFromISR(queues[3], &character, &xHigherPriorityTaskWoken);
			}
			break;
		case 3: //Receive line status
			if (LPC_UART3->LSR & (1 << 7)) {
				character = LPC_UART3->RBR;
			}
			break;
		case 4: //Character time out indicator
		default:
			character = LPC_UART3->LSR;
			break;
	}

	portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
}

uint8_t UART_Write(uint8_t uartNum, uint8_t *string, uint8_t length)
{
	uint8_t i;

	if (uartNum > 3) {
		return 0;
	}

	xSemaphoreTake(uartWriteMutex[uartNum], portMAX_DELAY);
	for(i = 0; i < length; i++) {
		while(!UartTxEmpty[uartNum]);
		switch(uartNum) {
			case 0:
				//Bit_off(LPC_UART0->IER, 1);
				LPC_UART0->THR = *(string++);
				//Bit_on(LPC_UART0->IER, 1);
				break;
			case 1:
				LPC_UART1->THR = *(string++);
				break;
			case 2:
				LPC_UART2->THR = *(string++);
				break;
			case 3:
				LPC_UART3->THR = *(string++);
				break;
		}
		UartTxEmpty[uartNum] = 0;
	}
	xSemaphoreGive(uartWriteMutex[uartNum]);

	return 1;
}
