#include "ecu.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "LPC17xx.h"
#include <stdint.h>
#include "UART.h"
#include "debug.h"
#include "Timer.h"
#include "bit.h"

static uint8_t firstFastInit = 1, ecuLinkUp = 0;
static xQueueHandle receiveQueue;
static uint32_t lastResponseTs = 0;
static xSemaphoreHandle writeLock;

static uint8_t _Ecu_Init();

void Ecu_Setup()
{
	//Set ecu enable signal to high
	Bit_off(LPC_PINCON->PINSEL0, 3 << ECU_ENABLE_PIN * 2);
	Bit_on(LPC_GPIO0->FIODIR, 1 << ECU_ENABLE_PIN);
	Bit_on(LPC_GPIO0->FIOSET, 1 << ECU_ENABLE_PIN);

	vSemaphoreCreateBinary(writeLock);
	Timer_Setup();
	receiveQueue = xQueueCreate(ECU_RESPONSE_MAX_FRAME_LENGTH, sizeof(uint8_t));
}

/**
 * Calculates the checksum of the given string as per ISO 14230 spec.
 *
 * @param string A pointer to the string to perform the checksum on
 * @param length The length of such string
 *
 * @return The checksum result
 */
static uint8_t _Ecu_CalcChecksum(uint8_t *string, uint8_t length)
{
	uint8_t checksum = 0, i;
	for(i = 0; i < length; i++) {
		checksum += *(string + i);
	}
	return checksum;
}

/**
 * Listens for an ECU response frame and parses it
 *
 * @return The raw ecu payload response
 */
static Ecu_Response _Ecu_GetResponse()
{
	static const responseGraph[] = {0x80, 0xF1, 0x11};
	uint8_t responseState = ECU_GRAPH_HEADER_0, responseExpectedLength = 0;
	uint8_t receivedString[ECU_RESPONSE_MAX_FRAME_LENGTH];
	uint8_t receivedLength = 0, inChar;
	Ecu_Response response = { .length=0, .success=0 };

	//First char response
	if (xQueueReceive(receiveQueue, &inChar, ECU_MAX_RESPONSE_DELAY)) {
		do {
			if (receivedLength == ECU_RESPONSE_MAX_FRAME_LENGTH) { //Frame too big
				DEBUG_EXPRESSION(printf("Received too large ECU frame\n"));
				break;
			}

			receivedString[receivedLength++] = inChar;
			lastResponseTs = *Timer_usTime;

			//We're now entering the response frame state graph
			switch(responseState) {
				//Check first 3 characters match the expected and fixed outcome
				case ECU_GRAPH_HEADER_0:
				case ECU_GRAPH_HEADER_1:
				case ECU_GRAPH_HEADER_2:
					if (inChar != responseGraph[responseState]) {
						responseState = ECU_GRAPH_ERROR;
					} else {
						++responseState;
					}
					break;
				//Get the expected payload length
				case ECU_GRAPH_LENGTH:
					if (inChar > ECU_RESPONSE_MAX_PAYLOAD_LENGTH) {
						responseState = ECU_GRAPH_ERROR;
					} else {
						responseExpectedLength = inChar;
						responseState = responseExpectedLength ? ECU_GRAPH_PAYLOAD : ECU_GRAPH_CHECKSUM;
					}
				//Read the payload characters
				case ECU_GRAPH_PAYLOAD:
					if (--responseExpectedLength == 0) {
						responseState = ECU_GRAPH_CHECKSUM;
					}
					response.message[response.length] = inChar;
					++(response.length);
					break;
				//Check if the checksum matches and return the response either way
				case ECU_GRAPH_CHECKSUM:
					if (_Ecu_CalcChecksum(receivedString, receivedLength) == inChar) {
						response.success = 1;
					}
					return response;
				//An error state, just for consuming the remaining bytes
				case ECU_GRAPH_ERROR:
					break;
			}
		} while(xQueueReceive(receiveQueue, &inChar, ECU_MAX_RESPONSE_INTER_BYTE_DELAY));
	} else {
		ecuLinkUp = 0;
		DEBUG_EXPRESSION(printf("ECU Response timed out\n"));
	}

	return response;
}

/**
 * The communication bus with the ECU is half-duplex, so in order to deplete
 * the receive buffer from the written data, we can use this function.
 * Also we check that the input matches with the expected char.
 *
 * @precondition A character has been written
 *
 * @param sentChar The character that we sent to the bus, and expect to receive back
 * @param waitTicks The maximum amount of ticks we'll wait for the echo
 *
 * @return 1 on success, 0 on failure
 */
static uint8_t _Ecu_WriteGetEcho(uint8_t sentChar, portTickType waitTicks)
{
	uint8_t inChar;

	if (!xQueueReceive(receiveQueue, &inChar, waitTicks)) {
		DEBUG_EXPRESSION(printf("Did not get write echo for 0x%x", sentChar));
	} else if (inChar != sentChar) {
		DEBUG_EXPRESSION(printf("Write echo missmatch, sent 0x%x got 0x%x", sentChar, inChar));
	} else {
		return 1;
	}

	return 0;
}

/**
 * Sends a request to the ECU and blocks until it gets a response.
 *
 * @param string The raw request frame W/O checksum
 * @param length The request length
 */
static Ecu_Response _Ecu_Write(uint8_t *string, uint8_t length)
{
	uint32_t usElapsed;
	uint8_t i, inChar, checkSum;
	Ecu_Response response;

	xSemaphoreTake(writeLock, portMAX_DELAY);
	while(xQueueReceive(receiveQueue, &inChar, 0)); //Empty queue

	//Delay write if not enough time has passed from the last request
	usElapsed = Timer_usElapsed(lastResponseTs);
	if (usElapsed < ECU_MIN_RESPONSE_REQUEST_DELAY * 1000) {
		vTaskDelay(ECU_MIN_RESPONSE_REQUEST_DELAY - usElapsed / 1000);
	}

	checkSum = _Ecu_CalcChecksum(string, length);
	//Slowly write to the UART as requested by ISO14230-2
	for(i=0; i < length; i++) {
		UART_Write(ECU_UART_NUM, string + i, 1);
		vTaskDelay(ECU_REQUEST_INTER_BYTE_DELAY);
		_Ecu_WriteGetEcho(*(string + i), 0);
	}

	UART_Write(ECU_UART_NUM, &checkSum, 1);
	_Ecu_WriteGetEcho(checkSum, ECU_REQUEST_INTER_BYTE_DELAY);

	response = _Ecu_GetResponse();
	xSemaphoreGive(writeLock);

	return response;
}

/**
 * Send a start communication request frame
 *
 * @return 0 on failure, 1 on success
 */
static uint8_t _Ecu_StartComm()
{
	uint8_t command[] = {0x81, 0x11, 0xF1, 0x81};
	Ecu_Response response = _Ecu_Write(command, 4);

	return response.success;
}

/**
 * Send a start diagnostic request frame
 *
 * @return 0 on failure, 1 on success
 */
static uint8_t _Ecu_StartDiagnostic()
{
	uint8_t command[] = {0x80, 0x11, 0xF1, 0x02, 0x10, 0x80};
	Ecu_Response response = _Ecu_Write(command, 6);

	return response.success;
}

/**
 * Send a fast init signal to the ECU, so that it starts listening
 * to commands.
 */
static void _Ecu_FastInit()
{
	portTickType ticks;

	Bit_off(LPC_PINCON->PINSEL0, 3 << 20);
	Bit_on(LPC_GPIO0->FIODIR, 1 << ECU_TX_PIN);
	Bit_on(LPC_GPIO0->FIOSET, 1 << ECU_TX_PIN);

	if (firstFastInit) { //We need to let the ECU initialize
		firstFastInit = 0;
		ticks = xTaskGetTickCount();
		if (ticks < ECU_INIT_DELAY) {
			vTaskDelay(ECU_INIT_DELAY - ticks + ECU_MIN_RESPONSE_REQUEST_DELAY);
		} else {
			vTaskDelay(ECU_MIN_RESPONSE_REQUEST_DELAY);
		}
	} else {
		vTaskDelay(ECU_MIN_RESPONSE_REQUEST_DELAY);
	}

	DEBUG_EXPRESSION(printf("ECU Fast init\n"));

	Bit_on(LPC_GPIO0->FIOCLR, 1 << ECU_TX_PIN);
	vTaskDelay(ECU_FASTINIT_PULSE_DURATION);
	Bit_on(LPC_GPIO0->FIOSET, 1 << ECU_TX_PIN);
	vTaskDelay(ECU_FASTINIT_PULSE_DURATION);
	UART_Setup(ECU_UART_NUM, UART_WORD_8_BIT, ECU_BAUD_RATE, receiveQueue);
}

/**
 * A wrapper function that takes all the steps required to enable the ECU
 * link
 *
 * @return 0 on failure, 1 on success
 */
static uint8_t _Ecu_Init()
{
	if (!ecuLinkUp) {
		_Ecu_FastInit();
		ecuLinkUp = _Ecu_StartComm() && _Ecu_StartDiagnostic();
	}

	return ecuLinkUp;
}

Ecu_Response Ecu_GetParameter(Ecu_ParameterType parameterAddress)
{
	uint8_t command[] = {0x80, 0x11, 0xF1, 0x02, 0x21, parameterAddress};
	uint8_t tries;
	Ecu_Response response = { .success = 0, .length = 0 };

	for(tries = ECU_GET_PARAMETER_TRIES; tries > 0; tries--) {
		if (ecuLinkUp || _Ecu_Init()) {
			response = _Ecu_Write(command, 6);
			if (response.success) {
				break;
			}
		}
	}

	return response;
}

int16_t Ecu_GetRpm()
{
	Ecu_Response response = Ecu_GetParameter(ECU_PARAM_RPM);
	if (response.success && response.length == 2) {
		return ECU_CALC_RPM(response.message);
	}

	return -1;
}

int16_t Ecu_GetThrottle()
{
	Ecu_Response response = Ecu_GetParameter(ECU_PARAM_THROTTLE);
	if (response.success && response.length == 2) {
		return ECU_CALC_THROTTLE(response.message);
	}

	return -1;
}
