/**
 * ECU module. Provides an abstraction layer to communicate to/from the
 * motorcycle ECU.
 */

#ifndef ECU_H_
#define ECU_H_

#include <stdint.h>

#define ECU_INIT_DELAY 300
#define ECU_GET_PARAMETER_TRIES 1

#define ECU_TX_PIN 10
#define ECU_RX_PIN 11
#define ECU_ENABLE_PIN 4

#define ECU_UART_NUM 2
#define ECU_BAUD_RATE 10400

#define ECU_FASTINIT_PULSE_DURATION 25

//Minimum delay between receiving a response and sending the next request
#define ECU_MIN_RESPONSE_REQUEST_DELAY 56

//Delay after a response times out
#define ECU_MAX_RESPONSE_DELAY 50

//Delay between each written byte to the ecu
#define ECU_REQUEST_INTER_BYTE_DELAY 10

//Delay after the next character of a response times out
#define ECU_MAX_RESPONSE_INTER_BYTE_DELAY 20

#define ECU_RESPONSE_MAX_PAYLOAD_LENGTH 4
#define ECU_RESPONSE_MAX_FRAME_LENGTH ECU_RESPONSE_MAX_PAYLOAD_LENGTH + 5

#define ECU_CALC_RPM(message) (message[0] * 100 + message[1])
#define ECU_CALC_THROTTLE(message) ((message[0] * 256 + message[1] - 0xD8) * 100 / (0x37F - 0xD8))

//Response graph states
#define ECU_GRAPH_HEADER_0 0
#define ECU_GRAPH_HEADER_1 1
#define ECU_GRAPH_HEADER_2 2
#define ECU_GRAPH_LENGTH 3
#define ECU_GRAPH_PAYLOAD 4
#define ECU_GRAPH_CHECKSUM 5
#define ECU_GRAPH_ERROR 9

typedef enum {
	ECU_PARAM_THROTTLE = 0x04,
	ECU_PARAM_COOLANT_TEMP = 0x06,
	ECU_PARAM_INTAKE_TEMP = 0x07,
	ECU_PARAM_RPM = 0x09,
	ECU_PARAM_GEAR = 0x0B,
	ECU_PARAM_SPEED = 0x0C
} Ecu_ParameterType;

typedef struct  {
	uint8_t message[ECU_RESPONSE_MAX_PAYLOAD_LENGTH];
	uint8_t length;
	uint8_t success;
} Ecu_Response;

/**
 * Initializes this ecu module
 */
void Ecu_Setup();

/**
 * Reads the RPMs from the ECU. This is a blocking call.
 * @precondition This is called from within a task
 *
 * @return The current RPM. -1 on failure
 */
int16_t Ecu_GetRpm();

/**
 * Reads the throttle opening angle % from the ECU. This is a blocking call.
 * @precondition This is called from within a task
 *
 * @return The percentage of the throttle opening angle. -1 on failure
 */
int16_t Ecu_GetThrottle();

/**
 * Reads a parameter from the ECU. This is a blocking call.
 * @precondition This is called from within a task
 *
 * @param The address of the parameter to read
 *
 * @return The raw ecu response payload
 */
Ecu_Response Ecu_GetParameter(Ecu_ParameterType parameterAddress);

#endif /* ECU_H_ */
