#include "ecuFeedback.h"
#include "ecu.h"
#include "FreeRTOS.h"
#include "task.h"
#include "lcd.h"
#include "log.h"

static int16_t averageRpm = 0;
static uint8_t numRpmSamples = 0;

void EcuFeedback_Setup()
{
	Ecu_Setup();
	Lcd_Setup();
}

/**
 * Updates the average RPM with the given new sample
 *
 * @param rpm The new rpm sample
 */
static void _EcuFeedback_addRpmSample(int16_t rpm) {
	averageRpm = (averageRpm * numRpmSamples + rpm) / (numRpmSamples + 1);
	if (numRpmSamples < ECUFEEDBACK_MAX_RPM_SAMPLES) {
		++numRpmSamples;
	}
}

/**
 * Determines the current driving behaviour(sport, standard or eco) and
 * makes the LCD backlight blink if a gear upshift is recommended.
 *
 * @param currentRpm The current engine RPM as obtained from the ECU
 */
static void _EcuFeedback_giveGearHint(int16_t currentRpm)
{
	int16_t thresholdRpm;

	if (averageRpm >= ECUFEEDBACK_SPORT_MODE_AVG) {
		thresholdRpm = ECUFEEDBACK_SPORT_MODE_SHIFT;
	} else if (averageRpm >= ECUFEEDBACK_STANDARD_MODE_AVG) {
		thresholdRpm = ECUFEEDBACK_STANDARD_MODE_SHIFT;
	} else {
		thresholdRpm = ECUFEEDBACK_ECO_MODE_SHIFT;
	}

	if (currentRpm >= thresholdRpm && !Lcd_isBlinking()) {
		Lcd_Blink(ECUFEEDBACK_BLINK_PERIOD, ECUFEEDBACK_BLINK_TIMES);
	}
}

void EcuFeedback_Task()
{
	int16_t rpm, throttle;
	uint8_t string[6];
	while(!Lcd_Acquire(portMAX_DELAY));
	Lcd_CursorPosition(0, 1);
	Lcd_Write("    0 RPM     0%");
	Lcd_Release();

	for(;;) {
		rpm = Ecu_GetRpm();
		throttle = Ecu_GetThrottle();

		if ((rpm >= 0 || throttle >= 0) && Lcd_Acquire(portMAX_DELAY)) {
			if (rpm >= 0) {
				Lcd_CursorPosition(0, 1);
				sprintf(string, "%5d", rpm);
				Lcd_Write(string);
				_EcuFeedback_addRpmSample(rpm);
				_EcuFeedback_giveGearHint(rpm);
			}

			if (throttle >= 0) {
				Lcd_CursorPosition(LCD_WIDTH - 4, 1);
				sprintf(string, "%3d", throttle);
				Lcd_Write(string);
			}

			Lcd_Release();
		} else if (LOG_IF_ENABLED(LOG_WARNING)) {
			Log_Write("Ecu feedback failed, RPM %d, Throttle %d", rpm, throttle);
		}

		vTaskDelay(ECUFEEDBACK_POLLING_INTERVAL);
	}
}
