#include "lcd.h"
#include "PWM.h"
#include "Timer.h"
#include "semphr.h"
#include "timers.h"

static uint8_t redChannel, greenChannel, blueChannel;
static lcd_rgb backlightColor = LCD_BLACK;
static xSemaphoreHandle lcdLock, blinkLock;
static uint8_t initialized = 0;
static xTimerHandle blinkTimer;
static uint32_t blinkTimes = 0;
static uint32_t blinkStartedOff = 0;

static void _Lcd_BlinkCallback(xTimerHandle timer);

void Lcd_Setup()
{
	if (initialized) {
		return;
	}

	initialized = 1;
	vSemaphoreCreateBinary(lcdLock);
	vSemaphoreCreateBinary(blinkLock);
	blinkTimer = xTimerCreate("LCDBlink", 100, 1, NULL, (tmrTIMER_CALLBACK) _Lcd_BlinkCallback);

	UART_Setup(LCD_UART_NUM, UART_WORD_8_BIT, LCD_UART_BAUDS, NULL);

	Timer_Setup();
	Timer_usDelay(LCD_POWER_UP_DELAY_US);

	redChannel = LCD_RED_CHANNEL;
	greenChannel = LCD_GREEN_CHANNEL;
	blueChannel = LCD_BLUE_CHANNEL;
	Lcd_BacklightOff();
	Lcd_ClearScreen();
	Lcd_on();
}

/**
 * Sets the backlight color
 *
 * @param color The RGB values
 */
static void _Lcd_SetBacklightColor(lcd_rgb *color)
{
	PWM_channelDuty(redChannel, color->red);
	PWM_channelDuty(greenChannel, color->green);
	PWM_channelDuty(blueChannel, color->blue);
	PWM_updateDuty();
}

void Lcd_SetBacklightColor(lcd_rgb color)
{
	backlightColor = color;
	_Lcd_SetBacklightColor(&color);
}


void Lcd_BacklightOff()
{
	static lcd_rgb black = LCD_BLACK;
	_Lcd_SetBacklightColor(&black);
}

void Lcd_BacklightOn()
{
	_Lcd_SetBacklightColor(&backlightColor);
}

void Lcd_on()
{
	Lcd_Write("\xFE\x41");
}

void Lcd_off()
{
	Lcd_Write("\xFE\x42");
}

void Lcd_CursorPosition(uint8_t col, uint8_t row)
{
	// col from 0 to 63
	uint8_t command[3] = { 0xFE, 0x45 };
	uint8_t position = 0;
	if (row == 1) {
		position = 0x40;
	}
	command[2] = position + col;
	UART_Write(0, command, 3);
}

void Lcd_CursorUnderlineOn()
{
	Lcd_Write("\xFE\x47");
}

void Lcd_CursorUnderlineOff()
{
	Lcd_Write("\xFE\x48");
}

void Lcd_CursorLeft()
{
	Lcd_Write("\xFE\x49");
}

void Lcd_CursorRight()
{
	Lcd_Write("\xFE\x4A");
}

void Lcd_CursorBlinkOn()
{
	Lcd_Write("\xFE\x4B");
}

void Lcd_CursorBlinkOff()
{
	Lcd_Write("\xFE\x4C");
}

void Lcd_Backspace()
{
	Lcd_Write("\xFE\x4E");
}

void Lcd_ClearScreen()
{
	Lcd_Write("\xFE\x51");
}

void Lcd_SetContrast(uint8_t contrast)
{
	uint8_t command[3] = { 0xFE, 0x52 };
	if (contrast >= 1 && contrast <= 50) {
		command[2] = contrast;
		UART_Write(0, command, 3);
	}
}

void Lcd_DisplayLeft()
{
	Lcd_Write("\xFE\x55");
}

void Lcd_DisplayRight()
{
	Lcd_Write("\xFE\x56");
}

/**
 * Calculates the length of a null terminated string
 *
 * @param string The string to check the length of
 *
 * @return The length of the string, W/O the null char
 */
static uint8_t _Lcd_Strlen(uint8_t *string)
{
	uint8_t length = 0;

	while(*(string + length) != 0) {
		++length;
	}

	return length;
}

uint8_t Lcd_Acquire(portBASE_TYPE blockTime)
{
	return xSemaphoreTake(lcdLock, blockTime) == pdTRUE;
}

void Lcd_Release()
{
	xSemaphoreGive(lcdLock);
}

void Lcd_Write(uint8_t *string)
{
	UART_Write(LCD_UART_NUM, string, _Lcd_Strlen(string));
}

/**
 * Software timer callback for the blink functionality. Should not be called
 * directly. Switches the backlight on or off.
 *
 * @param timer The software timer handle
 *
 */
static void _Lcd_BlinkCallback(xTimerHandle timer)
{
	if (blinkTimes == 0) {
		return Lcd_BlinkCancel();
	}

	if (blinkTimes-- & 1) {
		Lcd_BacklightOn();
	} else {
		Lcd_BacklightOff();
	}

}

/**
 * Stops the blink software timer if it's active
 */
static void _Lcd_BlinkCancel()
{
	if (!xTimerIsTimerActive(blinkTimer)) {
		return;
	}

	xTimerStop(blinkTimer, portMAX_DELAY);
	if (blinkStartedOff) {
		Lcd_BacklightOff();
	} else {
		Lcd_BacklightOn();
	}
}

uint8_t Lcd_isBlinking()
{
	return (uint8_t) xTimerIsTimerActive(blinkTimer);
}

void Lcd_BlinkCancel()
{
	xSemaphoreTake(blinkLock, portMAX_DELAY);
	_Lcd_BlinkCancel();
	xSemaphoreGive(blinkLock);
}

void Lcd_Blink(uint16_t period, uint32_t times)
{
	static lcd_rgb fallbackColor = LCD_BLUE;

	xSemaphoreTake(blinkLock, portMAX_DELAY);
	_Lcd_BlinkCancel();

	blinkStartedOff = backlightColor.blue == 0 && backlightColor.green == 0 && backlightColor.red == 0;
	if (blinkStartedOff) {
		Lcd_SetBacklightColor(fallbackColor);
	}

	blinkTimes = times * 2;
	xTimerChangePeriod(blinkTimer, period, portMAX_DELAY);
	xTimerStart(blinkTimer, portMAX_DELAY);
	xSemaphoreGive(blinkLock);
}
