#ifndef LCD_H_
#define LCD_H_

#include <stdint.h>
#include "UART.h"

typedef struct {
	uint8_t red;
	uint8_t green;
	uint8_t blue;
} lcd_rgb;

#define LCD_YELLOW { .red=100, .green=15, .blue=0 }
#define LCD_RED { .red=100, .green=0, .blue=0 }
#define LCD_GREEN { .red=0, .green=100, .blue=0 }
#define LCD_BLUE { .red=0, .green=0, .blue=100 }
#define LCD_WHITE { .red=100, .green=100, .blue=100 }
#define LCD_BLACK { .red=0, .green=0, .blue=0 }

#define LCD_POWER_UP_DELAY_US 100 * 1000
#define LCD_WIDTH 16

#define LCD_RED_CHANNEL 1
#define LCD_GREEN_CHANNEL 2
#define LCD_BLUE_CHANNEL 3

#define LCD_UART_NUM 0
#define LCD_UART_WORD_LENGTH UART_WORD_8_BIT
#define LCD_UART_BAUDS 9600

/**
 * Initializes the LCD module
 */
void Lcd_Setup();

/**
 * Sets the backlight color, but also retains it. This way we can switch
 * the backlight off, then on to the previous color.
 *
 * @param color The RGB values
 */
void Lcd_SetBacklightColor(lcd_rgb color);

/**
 * Turns off the backlight
 */
void Lcd_BacklightOff();

/**
 * Turns on the backlight
 */
void Lcd_BacklightOn();

/**
 * Enables the LCD display
 */
void Lcd_on();

/**
 * Disables the LCD display
 */
void Lcd_off();

/**
 * Sets the cursor position
 *
 * @param col The column position. Viewable columns are from 0 to 15.
 * @param row The row number, either 0 or 1
 */
void Lcd_CursorPosition(uint8_t col, uint8_t row);

/**
 * Shows the cursor position
 */
void Lcd_CursorUnderlineOn();

/**
 * Hids the cursor position
 */
void Lcd_CursorUnderlineOff();

/**
 * Moves the cursor one position to the left
 */
void Lcd_CursorLeft();

/**
 * Moves the cursosr one position to the right
 */
void Lcd_CursorRight();

/**
 * Makes the cursor blink(if visible)
 */
void Lcd_CursorBlinkOn();

/**
 * Disables the cursor blinking
 */
void Lcd_CursorBlinkOff();

/**
 * Delete the character before the current position, and move one position to the left
 */
void Lcd_Backspace();

/**
 * Clears the screen
 */
void Lcd_ClearScreen();

/**
 * Sets the constrast level of the screen
 *
 * @param contrast A value from 0 to 50, 50 being the most contrast
 */
void Lcd_SetContrast(uint8_t contrast);

/**
 * Shift the viewable area to the left
 */
void Lcd_DisplayLeft();

/**
 * Shift the viewable area to the right
 */
void Lcd_DisplayRight();

/**
 * Writes a string in the current cursor position
 * @param string A null terminated string to write
 */
void Lcd_Write(uint8_t *string);

/**
 * Acquires an exclusive use lock for the LCD, so that operations are atomic.
 * It's not enforced, so only works between callers to this function.
 *
 * @param blockTime The amount of ticks the caller can wait for acquiring the lock
 *
 * @return 0 on failure, 1 on success
 */
uint8_t Lcd_Acquire(portBASE_TYPE blockTime);

/**
 * Releases the lock for the LCD usage
 * @precondition The caller has previously successfully acquired a lock
 */
void Lcd_Release();

/**
 * Checks whether the blinking timer is active
 *
 * @return 1 if running, 0 otherwise
 */
uint8_t Lcd_isBlinking();

/**
 * Makes the LCD blink
 *
 * @param period Duration of each on/off state
 * @param times Number of blinks(on then off)
 */
void Lcd_Blink(uint16_t period, uint32_t times);

/**
 * Stops the blinking if it's active
 */
void Lcd_BlinkCancel();

#endif /* LCD_H_ */
