#include "log.h"
#include "FreeRTOS.h"
#include "task.h"
#include "UART.h"
#include "portmacro.h"
#include "semphr.h"

static uint8_t isLogEnabled = 0;
static uint8_t logLevel = LOG_DEFAULT_LEVEL;

static uint8_t buffer[LOG_BUFFER_SIZE];
static uint8_t bufferLength = 0;
static uint8_t uartChannel = 0;

static xSemaphoreHandle writeMutex;

void Log_Setup(uint8_t uartNum)
{
	uartChannel = uartNum;
	vSemaphoreCreateBinary(writeMutex);
}

void Log_Task()
{
	uint8_t writingLength;
	for(;;) {
		if (bufferLength > 0 && xSemaphoreTake(writeMutex, 0)) {
			UART_Write(uartChannel, buffer, bufferLength);
			bufferLength = 0;
			xSemaphoreGive(writeMutex);
		}

		vTaskDelay(5);
	}
}

void Log_Write(uint8_t *format, ...)
{
	va_list args;
	uint16_t length;

	xSemaphoreTake(writeMutex, portMAX_DELAY);
	va_start(args, format);
	length = vsnprintf(buffer + bufferLength, LOG_BUFFER_SIZE - bufferLength, format, args);

	//If not enough buffer to write, write from the start
	if (length <= LOG_BUFFER_SIZE && length > LOG_BUFFER_SIZE - bufferLength) {
		length = vsnprintf(buffer, LOG_BUFFER_SIZE, format, args);
		bufferLength = length;
	} else {
		bufferLength += length;
	}
	va_end(args);

	if (bufferLength < LOG_BUFFER_SIZE) {
		buffer[bufferLength] = '\r';
		++bufferLength;
		if (bufferLength < LOG_BUFFER_SIZE) {
			buffer[bufferLength] = '\n';
			++bufferLength;
		}
	}
	xSemaphoreGive(writeMutex);
}

Log_Level Log_SetLevel(Log_Level level)
{
	Log_Level previous = logLevel;
	logLevel = level;
	return previous;
}

uint8_t Log_Enabled(Log_Level level)
{
	return isLogEnabled && level >= logLevel;
}

void Log_Disable()
{
	isLogEnabled = 0;
}

void Log_Enable()
{
	isLogEnabled = 1;
}
