#include "FreeRTOS.h"
#include "proximity.h"
#include "queue.h"
#include "portmacro.h"
#include "debug.h"
#include "lcd.h"
#include "log.h"

static xQueueHandle measurementsQueue;
static lcd_rgb riskColors[] = { LCD_WHITE, LCD_GREEN, LCD_YELLOW, LCD_RED};

typedef enum {RISK_NONE, RISK_LOW, RISK_MEDIUM, RISK_HIGH} risk;

xQueueHandle Proximity_setup()
{
	Lcd_Setup();
	measurementsQueue = xQueueCreate(PROXIMITY_MEASUREMENTS_QUEUE_SIZE, sizeof(Sonar_measurement));
	return measurementsQueue;
}

/**
 * Outputs feedback for the given risk from the left positioned sonar
 *
 * @param risk The assesed risk level to output
 */
static void _Proximity_PrintLeftSonarRisk(risk printedRisk)
{
	uint8_t string[] = "<<<<", pos, i;

	for(pos = 3, i = 0; i < printedRisk; i++) {
		string[pos--] = ' ';
	}

	if (Lcd_Acquire(PROXIMITY_MS_PER_SAMPLE * 2 / 3)) {
		Lcd_CursorPosition(0, 0);
		Lcd_Write(string);
		Lcd_Release();
	}
}

/**
 * Outputs feedback for the given risk from the right positioned sonar
 *
 * @param risk The assesed risk level to output
 */
static void _Proximity_PrintRightSonarRisk(risk printedRisk)
{
	uint8_t string[] = ">>>>", pos;

	for(pos = 0; pos < printedRisk; pos++) {
		string[pos] = ' ';
	}

	if (Lcd_Acquire(PROXIMITY_MS_PER_SAMPLE  * 2 / 3)) {
		Lcd_CursorPosition(LCD_WIDTH - 4, 0);
		Lcd_Write(string);
		Lcd_Release();
	}
}

/**
 * Determine if the assessed risk warrants a feedback update or not, and
 * notifies the needed outputs
 *
 * @param sourceIndex The source the assessed risk comes from, 0 or 1
 * @param riskIndex The assessed risk index
 */
static void _Proximity_updateRiskFeedback(uint8_t sourceIndex, uint8_t riskIndex)
{
	static risk lastSourceRisk[] = {RISK_NONE, RISK_NONE}, lastRisk;
	risk currentRisk, newRisk;

	currentRisk = riskIndex > PROXIMITY_HIGH_SCORE_FROM ? RISK_HIGH :
			riskIndex > PROXIMITY_MEDIUM_SCORE_FROM ? RISK_MEDIUM :
			riskIndex > PROXIMITY_LOW_SCORE_FROM ? RISK_LOW : RISK_NONE;

	if (currentRisk != lastSourceRisk[sourceIndex]) {
		if (sourceIndex == PROXIMITY_LEFT_SONAR_SOURCE) {
			_Proximity_PrintLeftSonarRisk(currentRisk);
		} else {
			_Proximity_PrintRightSonarRisk(currentRisk);
		}
	}

	if (currentRisk < lastSourceRisk[(sourceIndex + 1) % 2]) {
		newRisk = lastSourceRisk[(sourceIndex + 1) % 2];
	} else {
		newRisk = currentRisk;
	}

	if (lastRisk != newRisk) {
		//Update backlight
		Lcd_SetBacklightColor(riskColors[newRisk]);
		DEBUG_EXPRESSION(printf("Setting new risk to %u\n", newRisk));
		if (LOG_IF_ENABLED(LOG_INFO)) {
			Log_Write("Updating risk to %u", newRisk);
		}
		lastRisk = newRisk;
	}

	lastSourceRisk[sourceIndex] = currentRisk;
}

void Proximity_dispatchTask()
{
	Sonar_measurement prevMeasurement[PROXIMITY_NUM_SOURCES], newMeasurement;
	uint8_t sourceIndex, speedRiskIndex, proximityRiskIndex, riskIndex;
	int32_t timeToImpactMs;
	for(;;) {
		if (!xQueueReceive(measurementsQueue, &newMeasurement, portMAX_DELAY)) {
			continue;
		}
		proximityRiskIndex = speedRiskIndex = 0;
		sourceIndex = newMeasurement.source;
		if (newMeasurement.distance >= 15 && newMeasurement.distance <= 375) {
			proximityRiskIndex = (375 - newMeasurement.distance) * 100 / (375 - 15);
		}

		if (&prevMeasurement[sourceIndex]) {
			//centimeters/second. Positive towards sensor, negative away from sensor
			timeToImpactMs = newMeasurement.distance * PROXIMITY_MS_PER_SAMPLE / (prevMeasurement[sourceIndex].distance - newMeasurement.distance);
			if (timeToImpactMs >= 300 && timeToImpactMs <= 3000) {
				speedRiskIndex = (-(timeToImpactMs - 300) * (timeToImpactMs - 300) / 72900) + 100;
			}
		}

		riskIndex = (speedRiskIndex > proximityRiskIndex) ? speedRiskIndex : proximityRiskIndex;
		DEBUG_EXPRESSION(printf("Proximity risk %u: %u, Speed risk: %u\n", sourceIndex, proximityRiskIndex, speedRiskIndex));

		_Proximity_updateRiskFeedback(sourceIndex, riskIndex);
		prevMeasurement[sourceIndex] = newMeasurement;
	}
}
