#include "FreeRTOS.h"
#include "GPIOINT.h"
#include "sonar.h"
#include <bit.h>
#include "task.h"
#include "LPC17xx.h"
#include "FreeRTOSConfig.h"
#include <limits.h>
#include "debug.h"
#include "Timer.h"

static GPIOINT_desc sonar1_rising, sonar1_falling, sonar2_rising, sonar2_falling;

static xQueueHandle measurementsQueue;

void Sonar_setup(xQueueHandle queue)
{
	Timer_Setup();
	measurementsQueue = queue;
	Bit_on(LPC_GPIO0->FIODIR, SONAR_FIO_PIN_MASK);
	Bit_on(LPC_GPIO0->FIOCLR, SONAR_FIO_PIN_MASK);

	GPIOINT_subscribe(&sonar1_rising, SONAR_RX_PORT, SONAR_RX_PIN1, rising, SONAR_QUEUE_SIZE);
	GPIOINT_subscribe(&sonar1_falling, SONAR_RX_PORT, SONAR_RX_PIN1, falling, SONAR_QUEUE_SIZE);
	GPIOINT_subscribe(&sonar2_rising, SONAR_RX_PORT, SONAR_RX_PIN2, rising, SONAR_QUEUE_SIZE);
	GPIOINT_subscribe(&sonar2_falling, SONAR_RX_PORT, SONAR_RX_PIN2, falling, SONAR_QUEUE_SIZE);

	GPIOINT_enable();
}

void Sonar_triggerTask()
{
	volatile uint8_t i;
	for(;;) {
		//Trigger the sonar 1
		Bit_on(LPC_GPIO0->FIOSET, 1 << SONAR_TRIGGER_PIN1);
		Timer_usDelay(11);
		Bit_on(LPC_GPIO0->FIOCLR, 1 << SONAR_TRIGGER_PIN1);
		vTaskDelay(40);

		//Trigger sonar 2
		Bit_on(LPC_GPIO0->FIOSET, 1 << SONAR_TRIGGER_PIN2);
		Timer_usDelay(11);
		Bit_on(LPC_GPIO0->FIOCLR, 1 << SONAR_TRIGGER_PIN2);

		if (SONAR_TRIGGER_DELAY > 40) {
			vTaskDelay(SONAR_TRIGGER_DELAY - 40);
		}
	}
}

/**
 * Transforms a measured tick interval into a distance
 *
 * @param from The ticks measured at the first interrupt
 * @param to The ticks measured at the second interrupt
 * @return The measured distance in centimeters
 */
static uint16_t _Sonar_calcDistance(uint32_t from, uint32_t to)
{
	uint32_t elapsed;
	if (to < from) { //Overflow
		elapsed = (ULONG_MAX - from + to);
	} else {
		elapsed = (to - from);
	}

	return elapsed * 17 / 1000;
}

void Sonar_distanceTaskDevice1()
{
	uint32_t from, to;
	Sonar_measurement measurement = { .source = 0 };
	for(;;) {
		xQueueReceive(sonar1_rising.queueHandle, &from, portMAX_DELAY);
		if (xQueueReceive(sonar1_falling.queueHandle, &to, SONAR_MEASUREMENT_MAX_DELAY)) {
			measurement.distance = _Sonar_calcDistance(from, to);
			DEBUG_EXPRESSION(printf("SONAR 1 Measured %u cm\n", measurement.distance));
			xQueueSendToBack(measurementsQueue, &measurement, SONAR_MEASUREMENTS_QUEUE_WAIT);
		}
	}
}

void Sonar_distanceTaskDevice2()
{
	uint32_t from, to;
	Sonar_measurement measurement = { .source = 1 };
	for(;;) {
		xQueueReceive(sonar2_rising.queueHandle, &from, portMAX_DELAY);
		if (xQueueReceive(sonar2_falling.queueHandle, &to, SONAR_MEASUREMENT_MAX_DELAY)) {
			measurement.distance = _Sonar_calcDistance(from, to);
			DEBUG_EXPRESSION(printf("SONAR 2 Measured %u cm\n", measurement.distance));
			xQueueSendToBack(measurementsQueue, &measurement, SONAR_MEASUREMENTS_QUEUE_WAIT);
		}
	}
}
